import numpy as np
from tqdm import tqdm
import scipy.io as sio
import os
import pkg_resources
import hickle

# cosmology assumption
from astropy.cosmology import FlatLambdaCDM
cosmo = FlatLambdaCDM(H0=70, Om0=0.3)

from .priors import *
from .gp_sfh import *

try:
    import fsps
    mocksp = fsps.StellarPopulation(compute_vega_mags=False, zcontinuous=1,sfh=0, imf_type=1, logzsol=0.0, dust_type=2, dust2=0.0, add_neb_emission=True)
    print('Starting dense_basis. please wait ~ a minute for the FSPS backend to initialize.')

except:
    mocksp = None
    print('Starting dense_basis. Failed to load FSPS, only GP-SFH module will be available.')

priors = Priors()

#-----------------------------------------------------------------------
#                     Calculating spectra and SEDs
#-----------------------------------------------------------------------

def get_path(filename):
    return os.path.dirname(os.path.realpath(filename))

def convert_to_microjansky(spec,z,cosmology):
    """Convert a spectrum in L_\nu (Solar luminosity/Hz, default python-fsps
        output) to F_\nu units in microjansky
        Args:
            spec[1d numpy array]: Spectrum output by FSPS (L_\nu)
            z[float]: redshift for computing luminosity_distance
            cosmo[astropy cosmology object]: cosmology
        Returns:
            spec[1d numpy array]: Spectrum in F_\nu (\muJy)
        """

    temp = (1+z)*spec *1e6 * 1e23*3.48e33/(4*np.pi*3.086e+24*3.086e+24*cosmo.luminosity_distance(z).value*cosmo.luminosity_distance(z).value)
    #temp = spec *1e6 * 1e23*3.48e33/(4*np.pi*3.086e+24*3.086e+24*cosmo.luminosity_distance(z).value*cosmo.luminosity_distance(z).value*(1+z))
    return temp

def makespec_atlas(atlas, galid, priors, sp, cosmo, filter_list = [], filt_dir = [], return_spec = False):

    sfh_tuple = atlas['sfh_tuple'][galid,0:]
    zval = atlas['zval'][galid]
    dust = atlas['dust'][galid]
    met = atlas['met'][galid]

    specdetails = [sfh_tuple, dust, met, zval]
    if priors.dynamic_decouple == True:
        priors.decouple_sfr_time = 100*cosmo.age(zval).value/cosmo.age(0.1).value

    output = makespec(specdetails, priors, sp, cosmo, filter_list, filt_dir, return_spec)

    return output

def makespec(specdetails, priors, sp, cosmo, filter_list = [], filt_dir = [], return_spec = False, peraa = False, input_sfh = False):
    """
    makespec() works in two ways to create spectra or SEDs from an input list of physical paramters.
    with input_sfh = False, specdetails = [sfh_tuple, dust, met, zval]
    with input_sfh = True, specdetails = [sfh, timeax, dust, met, zval]

    return_spec can be true, false, or an array of wavelengths. in case

    it uses the db.mocksp object for the underlying SPS calculations, so make sure it's set accordingly.
    it also uses the priors object for things like decouple_sfr.

    """


    # hardcoded parameters - offload this to a seprarate function
    sp.params['sfh'] = 3
    sp.params['cloudy_dust'] = True
    sp.params['gas_logu'] = -2
    sp.params['add_igm_absorption'] = True
    sp.params['add_neb_emission'] = True
    sp.params['add_neb_continuum'] = True
    sp.params['imf_type'] = 1 # Chabrier

    # variable parameters
    if input_sfh == True:
        [sfh, timeax, dust, met, zval] = specdetails
    else:
        [sfh_tuple, dust, met, zval] = specdetails
        sfh, timeax = tuple_to_sfh(sfh_tuple, zval, decouple_sfr = priors.decouple_sfr, decouple_sfr_time = priors.decouple_sfr_time)
    sp.set_tabular_sfh(timeax, sfh)
    # sp.params['dust_type'] = 2
    # sp.params['dust1'] = dust1_rand
    sp.params['dust2'] = dust
    sp.params['logzsol'] = met
    sp.params['gas_logz'] = met # matching stellar to gas-phase metallicity
    sp.params['zred'] = zval

    #lam, spec = sp.get_spectrum(tage = cosmo.age(zval).value, peraa = peraa)
    # adding 0.1 Myr here to get the last couple of FSPS SSPs
    lam, spec = sp.get_spectrum(tage = cosmo.age(zval).value+1e-4, peraa = peraa)
    spec_ujy = convert_to_microjansky(spec, zval, cosmo)

    if type(return_spec) == type(True):

        if return_spec == True:
            return lam, spec_ujy

        elif return_spec == False:
            filcurves, _, _ = make_filvalkit_simple(lam, zval, fkit_name = filter_list, filt_dir = filt_dir)
            sed = calc_fnu_sed_fast(spec_ujy, filcurves)
            return sed

    elif len(return_spec) > 10:
        return convert_to_splined_spec(spec, lam, return_spec, zval)

    else:
        raise('Unknown argument for return_spec. Use True or False, or pass a wavelength grid.')

    return 0

def convert_to_splined_spec(spec_peraa, lam, lam_spline, redshift, cosmology = cosmo):

    spec = spec_peraa
    spec_ergsec = spec*3.839e33*1e17/(1+redshift)
    lum_dist = cosmology.luminosity_distance(redshift).value
    spec_ergsec_cm2 = spec_ergsec/(4*np.pi*3.086e+24*3.086e+24*lum_dist*lum_dist)
    # this should have units of 1e-17 erg/s/Ang
    # need units to be flux in 1e-17 erg/s/cm^2/Ang/spaxel
    spec_spline = np.interp(lam_spline, lam*(1+redshift), spec_ergsec_cm2)

    #return spec_ergsec_cm2
    return spec_spline

def make_sed_fast(sfh_tuple, metval, dustval, zval, filcurves, igmval = True, return_lam = False, sp = mocksp, cosmology = cosmo):

    """Generate and multiply a spectrum with previously generated
        filter transmission curves to get SED.
        ---WARNING: assumes filcurves have been generated at the correct zval---
        Args:
            sfh_tuple[1d numpy array]: Spectrum output by FSPS (L_\nu)
            metval[float]: log metallicity wrt Solar.
            dustval[float]: Calzetti dust attenuation
            zval[float]: redshift
            filcurves[2d array, (len(spec), Nfilters)]: filter transmission
                curves splined to wavelength array. Generated using the
                make_filvalkit_simple function.
            igmval[float, optional]: Include IGM absorption (Default is True)
            return_lam[boolean, optional]: Return a wavelength array
                along with the spectrum (Default is True)
            sp[stellar population object]: FSPS stellar population object.
                Initialized previously for speed.
            cosmo[astropy cosmology object]: cosmology.
                Default is FlatLambdaCDM.
        Returns:
            sed [1d numpy array, len = Nfilters]: SED in F_nu (muJy)
        """

    spec, logsfr, logmstar = make_spec(sfh_tuple, metval, dustval, zval, igmval = True, return_ms = True, return_lam = False, sp = sp, cosmology = cosmo)
    sed = calc_fnu_sed_fast(spec, filcurves)
    return sed, logsfr, logmstar

def make_filvalkit_simple(lam,z, fkit_name = 'filter_list.dat' ,vb=False, filt_dir = 'filters/'):

    # import os
    # print(os.listdir())
    lam_z = (1+z)*lam

    # change this to logspace later to avoid problems
    # when dealing with FIR filters.
    lam_z_lores = np.arange(2000,150000,2000)

    if filt_dir == 'internal':
        resource_package = __name__
        resource_path = '/'.join(('filters', fkit_name))  # Do not use os.path.join()
        template = pkg_resources.resource_string(resource_package, resource_path)
        f = template.split()
        temp = template.split()
    else:
        if filt_dir[-1] == '/':
            f = open(filt_dir+fkit_name,'r')
        else:
            f = open(filt_dir+'/'+fkit_name,'r')
        temp = f.readlines()
        #resource_package = os.path.dirname(os.path.realpath(__file__))
        #resource_package = __file__
        # resource_package = __name__
        # if filt_dir[-1] == '/':
        #     resource_path = '/'.join((filt_dir[0:-1], fkit_name))
        # else:
        #     resource_path = '/'.join((filt_dir, fkit_name))  # Do not use os.path.join()
        #
        # print(resource_package)
        # print(resource_path)
        # template = pkg_resources.resource_string(resource_package, resource_path)
        # f = template.split()
        # temp = template.split()
        #
        # print(temp)

    # read in the file with the filter curves

    if vb == True:
        print('number of filters to be read in: '+str(len(temp)))

    numlines = len(temp)
    if temp[-1] == '\n':
        numlines = len(temp)-1

    filcurves = np.zeros((len(lam_z),numlines))
    filcurves_lores = np.zeros((len(lam_z_lores),numlines))

    if vb == True:
        plt.figure(figsize= (12,6))

    for i in range(numlines):

        if (filt_dir == 'internal') & (fkit_name == 'filter_list_goodss.dat'):
            tempfilt = np.loadtxt(get_file('filters/filter_curves/goods_s', temp[i][22:].decode("utf-8")))
        elif (filt_dir == 'internal') & (fkit_name == 'filter_list_goodsn.dat'):
            tempfilt = np.loadtxt(get_file('filters/filter_curves/goods_n', temp[i][22:].decode("utf-8")))
        elif (filt_dir == 'internal') & (fkit_name == 'filter_list_cosmos.dat'):
            tempfilt = np.loadtxt(get_file('filters/filter_curves/cosmos', temp[i][21:].decode("utf-8")))
        elif (filt_dir == 'internal') & (fkit_name == 'filter_list_egs.dat'):
            tempfilt = np.loadtxt(get_file('filters/filter_curves/egs', temp[i][18:].decode("utf-8")))
        elif (filt_dir == 'internal') & (fkit_name == 'filter_list_uds.dat'):
            tempfilt = np.loadtxt(get_file('filters/filter_curves/uds', temp[i][18:].decode("utf-8")))
        else:
            if filt_dir[-1] == '/':
                filt_name = filt_dir+temp[i]
            else:
                filt_name = filt_dir+'/'+temp[i]
            if i == numlines-1:
                #print(filt_name[0:][0:])
                tempfilt = np.loadtxt(filt_name[0:-1])
            else:
                if os.path.exists(filt_name[0:][0:-1]):
                    tempfilt = np.loadtxt(filt_name[0:][0:-1])
                else:
                    raise Exception('filters not found. are you sure the folder exists at the right relative path?')

        temp_lam_arr = tempfilt[0:,0]
        temp_response_curve = tempfilt[0:,1]

        bot_val = np.amin(np.abs(lam_z - np.amin(temp_lam_arr)))
        bot_in = np.argmin(np.abs(lam_z - np.amin(temp_lam_arr)))
        top_val = np.amin(np.abs(lam_z - np.amax(temp_lam_arr)))
        top_in = np.argmin(np.abs(lam_z - np.amax(temp_lam_arr)))

        curve_small = np.interp(lam_z[bot_in+1:top_in-1],temp_lam_arr,temp_response_curve)
        splinedcurve = np.zeros((lam_z.shape))
        splinedcurve[bot_in+1:top_in-1] = curve_small
        if np.amax(splinedcurve) > 1:
            splinedcurve = splinedcurve / np.amax(splinedcurve)

        filcurves[0:,i] = splinedcurve

        if vb == True:
            plt.plot(np.log10(lam_z),splinedcurve,'k--',label=filt_name[0:][0:-1])

    if (filt_dir != 'internal'):
        f.close()

    if vb == True:

        print('created filcurve array splined to input lambda array at redshift: '+str(z))
        plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        plt.xlabel(r'log $\lambda [\AA]$',fontsize=18)
        plt.ylabel('Filter transmission')
        plt.axis([3.5,5,0,1])
        plt.show()

    return filcurves, lam_z, lam_z_lores

def calc_fnu_sed(spec,z,lam, fkit_name = 'filter_list.dat', filt_dir = 'filters/'):

    #kitname = '/home/kiyer/Documents/codex_dense_basis_data/new_stoc_analysis/letter1_fkilt/codex_filter_val_kit_fsps_'+str(z)+'.mat'
    #filt_data = sio.loadmat(kitname)
    #lam_z = np.array(filt_data['lambda_z'])
    #lam_z_lores = np.array(filt_data['lambda_z_lores'])
    #lam_z_input = lam*(1+z)

    filcurves,lam_z, lam_z_lores = make_filvalkit_simple(lam,z,fkit_name = fkit_name, filt_dir = filt_dir)

    #spec_interp = np.interp(lam_z,lam_z_input,spec)

    nu = 3e18/lam_z

    # change this to appropriate normalisation. see documentation.
    #fnuspec = spec*768.3
    fnuspec = spec
    #print(lam_z_lores.shape,lam_z.shape,fnuspec.shape)
    #fnuspec_lores = np.interp(lam_z_lores,lam_z,fnuspec)

    #filcurves = np.array(filt_data['filtercurve'])
    filvals = np.zeros((filcurves.shape[1],))
    for tindex in range(filcurves.shape[1]):
        #print(tindex)
        temp1 = filcurves[np.where(filcurves[0:,tindex]>0),tindex]
        temp2 = fnuspec[np.where(filcurves[0:,tindex]>0)]
        filvals[tindex] = np.sum(temp1.T[0:,0]*temp2)/np.sum(filcurves[0:,tindex])
        #print(filvals[tindex])
        #filvals[tindex] = temp1.T[0:,0]*temp2*np.sum(filcurves[0:,tindex])
        #print(temp1.T.shape,temp2.shape,temp3.shape,np.sum(filcurves[0:,i]))
        #filvals[i] = filcurves[np.where(filcurves[0:,i]>0),i]*fnuspec[np.where(filcurves[0:,i]>0)]/np.sum(filcurves[0:,i])
    return filvals

#def calc_fnu_sed_fast(spec,z,lam,filcurves,lam_z, lam_z_lores):
def calc_fnu_sed_fast(fnuspec,filcurves):
    filvals = np.zeros((filcurves.shape[1],))
    for tindex in range(filcurves.shape[1]):
        temp1 = filcurves[np.where(filcurves[0:,tindex]>0),tindex]
        temp2 = fnuspec[np.where(filcurves[0:,tindex]>0)]
        filvals[tindex] = np.sum(temp1.T[0:,0]*temp2)/np.sum(filcurves[0:,tindex])
    return filvals

def generate_atlas(N_pregrid = 10, priors=priors, initial_seed = 42, store = True, filter_list = 'filter_list.dat', filt_dir = 'filters/', norm_method = 'median', z_step = 0.01, sp = mocksp, cosmology = cosmo, fname = None, path = 'pregrids/', lam_array_spline = [], rseed = None):

    """Generate a pregrid of galaxy properties and their corresponding SEDs
        drawn from the prior distributions defined in priors.py
        Args:
            N_pregrid[int]: Number of SEDs in the pre-grid.
            Nparam[int]: Number of SFH parameters in each tuple
            Initial_seed[int]: Initial seed for random number generation.
            store[Boolean, optional]: Flag whether to store results
                or return as output
            filter_list[filename]: File that contains a list of filter curves.
            z_step[float]: Step size in redshift for filter curve grid.
                Default is 0.01.
            norm_method[string, default = 'max']: normalization for SEDs and SFHs.
                Currently supported arguments are 'none', 'max', 'median', 'area'.
            sp[stellar population object]: FSPS stellar population object. Initialized previously for speed.
            cosmo[astropy cosmology object]: cosmology. Default is FlatLambdaCDM
        Returns:
            [if store == False]
            rand_sfh_tuples[2d numpy array]: N_samples prior-sampled SFH tuples
            rand_Z: prior-sampled metallicity values
            rand_Av: prior-sampled dust attenuation values
            rand_z: prior-sampled redshift values
            rand_seds: Corresponding SEDs in F_\nu (\muJy)
            norm_method: Argument for how SEDs are normalized, pass into fitter
        """

    print('generating atlas with: ')
    print(priors.Nparam, ' tx parameters, ', priors.sfr_prior_type, ' SFR sampling', priors.sfh_treatment,' SFH treatment', priors.met_treatment,' met sampling', priors.dust_model, ' dust attenuation', priors.dust_prior,' dust prior', priors.decouple_sfr,' SFR decoupling.')

    if rseed is not None:
        print('setting random seed to :',rseed)
        np.random.seed(rseed)

    zval_all = []
    sfh_tuple_all = []
    sfh_tuple_rec_all = []
    norm_all = []
    dust_all = []
    met_all = []
    sed_all = []
    mstar_all = []
    sfr_all = []
    sim_timeax_all = []
    sim_sfh_all = []

    Nparam = priors.Nparam

    for i in tqdm(range(int(N_pregrid))):

        zval = priors.sample_z_prior()

        massval = priors.sample_mass_prior()
        if priors.sfr_prior_type == 'sSFRlognormal':
            sfrval = priors.sample_sfr_prior(zval=zval)
        else:
            sfrval = priors.sample_sfr_prior()
        txparam = priors.sample_tx_prior()
        sfh_tuple = np.hstack((massval, sfrval, Nparam, txparam))
        norm = 1.0

        if priors.dynamic_decouple == True:
            priors.decouple_sfr_time = 100*cosmo.age(zval).value/cosmo.age(0.1).value
            #print('decouple time: %.2f myr at z: %.2f' %(priors.decouple_sfr_time,zval))
        sfh, timeax = tuple_to_sfh(sfh_tuple, zval, decouple_sfr = priors.decouple_sfr, decouple_sfr_time = priors.decouple_sfr_time)

        temp = calctimes(timeax, sfh, priors.Nparam)
        temptuple = calctimes_to_tuple(temp)

        dust = priors.sample_Av_prior()
        met = priors.sample_Z_prior()

        #-------------------------------------------

#         sp.params['sfh'] = 3
#         sp.set_tabular_sfh(timeax, sfh)
#         sp.params['dust2'] = dust
#         sp.params['logzsol'] = met
#         sp.params['add_igm_absorption'] = True
#         sp.params['add_neb_emission'] = True
#         sp.params['add_neb_continuum'] = True
#         sp.params['imf_type'] = 1 # Chabrier
#         sp.params['zred'] = zval

#         lam, spec = sp.get_spectrum(tage = cosmology.age(zval).value)
#         spec_ujy = convert_to_microjansky(spec, zval, cosmology)

        specdetails = [sfh_tuple, dust, met, zval]

        if len(lam_array_spline) > 0:
            sed = makespec(specdetails, priors, sp, cosmology, filter_list, filt_dir, return_spec = lam_array_spline, peraa = True)
        else:
            lam, spec_ujy = makespec(specdetails, priors, sp, cosmology, filter_list, filt_dir, return_spec = True)

            if i == 0:
                # make grid of filter transmission curves for faster computation
                fc_zgrid = np.arange(priors.z_min-z_step, priors.z_max+z_step, z_step)
                temp_fc, temp_lz, temp_lz_lores = make_filvalkit_simple(lam,priors.z_min,fkit_name = filter_list, filt_dir = filt_dir)

                fcs= np.zeros((temp_fc.shape[0], temp_fc.shape[1], len(fc_zgrid)))
                lzs = np.zeros((temp_lz.shape[0], len(fc_zgrid)))
                lzs_lores = np.zeros((temp_lz_lores.shape[0], len(fc_zgrid)))

                for i in (range(len(fc_zgrid))):
                    fcs[0:,0:,i], lzs[0:,i], lzs_lores[0:,i] = make_filvalkit_simple(lam,fc_zgrid[i],fkit_name = filter_list, filt_dir = filt_dir)

            fc_index = np.argmin(np.abs(zval - fc_zgrid))
            sed = calc_fnu_sed_fast(spec_ujy, fcs[0:,0:,fc_index])

        #-------------------------------------------

        if norm_method == 'none':
            # no normalization
            norm_fac = 1
        elif norm_method == 'max':
            # normalize SEDs to 1 - seems to work better than median for small grids
            norm_fac = np.amax(sed)
        elif norm_method == 'median':
            # normalize SEDs to median
            norm_fac = np.nanmedian(sed)
        elif norm_method == 'area':
            # normalize SFH to 10^9 Msun
            norm_fac == 10**(massval - 9)
        else:
            raise ValueError('undefined normalization argument')

        sed = sed/norm_fac
        norm = norm/norm_fac
        mstar = np.log10(sp.stellar_mass / norm_fac)
        sfr = np.log10(sp.sfr / norm_fac)
        sfh_tuple[0] = sfh_tuple[0] - np.log10(norm_fac)
        sfh_tuple[1] = sfh_tuple[1] - np.log10(norm_fac)
        temptuple[0] = temptuple[0] - np.log10(norm_fac)
        temptuple[1] = temptuple[1] - np.log10(norm_fac)

        #-------------------------------------------

        zval_all.append(zval)
        sfh_tuple_all.append(sfh_tuple)
        sfh_tuple_rec_all.append(temptuple)
        norm_all.append(norm)
        dust_all.append(dust)
        met_all.append(met)
        sed_all.append(sed)
        mstar_all.append(mstar)
        sfr_all.append(sfr)


    pregrid_dict = {'zval':np.array(zval_all),
                   'sfh_tuple':np.array(sfh_tuple_all),
                   'sfh_tuple_rec':np.array(sfh_tuple_rec_all),
                   'norm':np.array(norm_all), 'norm_method':norm_method,
                   'mstar':np.array(mstar_all), 'sfr':np.array(sfr_all),
                   'dust':np.array(dust_all), 'met':np.array(met_all),
                   'sed':np.array(sed_all)}

    if store == True:

        if fname is None:
            fname = 'sfh_pregrid_size'
        if os.path.exists(path):
            print('Path exists. Saved atlas at : '+path+fname+'_'+str(N_pregrid)+'_Nparam_'+str(Nparam)+'.dbatlas')
        else:
            os.mkdir(path)
            print('Created directory and saved atlas at : '+path+fname+'_'+str(N_pregrid)+'_Nparam_'+str(Nparam)+'.dbatlas')
        try:
            hickle.dump(pregrid_dict,
                        path+fname+'_'+str(N_pregrid)+'_Nparam_'+str(Nparam)+'.dbatlas',
                        compression='gzip', compression_opts = 9)
        except:
            print('storing without compression')
            hickle.dump(pregrid_dict,
                        path+fname+'_'+str(N_pregrid)+'_Nparam_'+str(Nparam)+'.dbatlas')

        return

    return pregrid_dict

def load_atlas(fname, N_pregrid, N_param, path = 'pregrids/'):

    fname_full = path+fname+'_'+str(N_pregrid)+'_Nparam_'+str(N_param)+'.dbatlas'
    cat = hickle.load(fname_full)
    return cat

def quantile_names(N_params):
    return (np.round(np.linspace(0,100,N_params+2)))[1:-1]


#---------------- deprecated functions -----------------------

# def make_spec_deprecated(sfh_tuple, metval, dustval, zval, igmval = True, return_lam = False, return_ms = False, sp = mocksp, cosmology = cosmo):

#     """Use FSPS to generate a spectrum corresponding to a set of
#         input galaxy properties.
#         Args:
#             sfh_tuple[1d numpy array]: SFH parameters, input to gp_sfh_sklearn
#             metval[float]: log metallicity wrt Solar.
#             dustval[float]: Calzetti dust attenuation
#             zval[float]: redshift
#             igmval[float, optional]: Include IGM absorption (Default is True)
#             return_lam[boolean, optional]: Return a wavelength array along
#                 with the spectrum (Default is True)
#             sp[stellar population object]: FSPS stellar population object.
#                 Initialized previously for speed.
#             cosmo[astropy cosmology object]: cosmology.
#                 Default is FlatLambdaCDM
#         Returns:
#             spec[1d numpy array]: Spectrum in F_\nu (\muJy)
#             lam[1d numpy array]: Wavelength in Angstrom corresponding to spectrum
#         """

#     sp.params['gas_logu'] = -2
#     sp.params['sfh'] = 3
#     sp.params['cloudy_dust'] = True
#     sp.params['dust_type'] = 2

#     sp.params['add_igm_absorption'] = igmval
#     sp.params['zred'] = zval

#     sfh, timeax = tuple_to_sfh(sfh_tuple, zval = zval)
#     timeax = timeax
#     sp.set_tabular_sfh(timeax, sfh)

#     # sp.params['dust1'] = dust1_rand
#     sp.params['dust2'] = dustval
#     sp.params['logzsol'] = metval
#     [lam_arr,spec] = sp.get_spectrum(tage = np.amax(timeax))
#     spec_ujy = convert_to_microjansky(spec,zval,cosmology)

#     # add option to work in energy or F_\lambda as well

#     if return_ms == True:
#         return spec_ujy, np.log10(sfh[-1]), np.log10(sp.stellar_mass)
#     elif return_lam == True:
#         return spec_ujy, lam_arr
#     else:
#         return spec_ujy



# SED fitting priors and modeling assumptions, all in one place

import numpy as np
import matplotlib.pyplot as plt
import corner

# cosmology assumption
from astropy.cosmology import FlatLambdaCDM
cosmo = FlatLambdaCDM(H0=70, Om0=0.3)

from .gp_sfh import *

def quantile_names(N_params):
    return (np.round(np.linspace(0,100,N_params+2)))[1:-1]

class Priors(object):
    """ A class the holds prior information for the various parameters
    during SED fitting - their distributions and bounds. Gets passed into
    generate_atlas methods.

    Attributes
    ----------
    mass : log stellar mass
        upper and lower limits of distribution
    sfr : log instantaneous SFR
        upper and lower limits of distribution
    ssfr : log specific SFR = SFR/M*
        negative lognormal distribution parameters
    tx : lookback times at which a galaxy formed certain fractions of its mass
        concentration (tx_alpha) and number of parameters (Nparam) of distribution
    Z  : log metallicity/solar
        upper and lower limits of distribution
    Av : calzetti dust attenuation
        upper and lower limits of distribution
    z  : redshift
        upper and lower limits of distribution

    Methods
    -------
    sample_mass_prior : sample from mass prior
    sample_sfr_prior : sample from SFR prior
    sample_tx_prior : sample from tx prior
    sample_Av_prior : sample from dust attenuation prior
    sample_Z_prior : sample from metallicity prior
    sample_z_prior : sample from redshift prior
    print_priors : print active priors and their distributions
    sample_sfh_tuple : sample SFH priors
    sample_all_params : sample all priors
    sample_all_params_safesSFR : sample all priors with an sSFR limit to ensure gp_sfh doesn't have negative derivatives
    make_N_prior_draws : draw N parameter tuples from prior space
    plot_prior_distributions : plot corner plot of parameters and their priors
    plot_sfh_prlor : compute and plot the effective prior in SFH space

    """

    def __init__(self):
        self.mass_max = 12.0
        self.mass_min = 9.0

        self.sfr_prior_type = 'sSFRflat' # options are SFRflat, sSFRflat, sSFRlognormal
        self.sfr_max = -1.0
        self.sfr_min = 2.0
        self.ssfr_min = -12.0
        self.ssfr_max = -7.5
        self.ssfr_mean = 0.6
        self.ssfr_sigma = 0.4 # roughly ~2.5 dex width, use 0.1 for a tighter correlation
        self.ssfr_shift = -0.3

        self.z_min = 0.9
        self.z_max = 1.1

        self.met_treatment = 'flat' # options are flat and massmet
        self.Z_min = -1.5
        self.Z_max = 0.25
        self.massmet_width = 0.3

        self.dust_model = 'Calzetti'
        self.dust_prior = 'exp'
        self.Av_min = 0.0
        self.Av_max = 1.0
        self.Av_exp_scale = 1.0/3.0

        self.sfh_treatment = 'custom' # options are custom and TNGlike
        self.tx_alpha = 5.0
        self.Nparam = 3
        self.decouple_sfr = False
        self.decouple_sfr_time = 100 # in Myr
        self.dynamic_decouple = True # set decouple time according to redshift (100 Myr at z=0.1)


    def sample_mass_prior(self, size = 1):
        massval = np.random.uniform(size=size)*(self.mass_max-self.mass_min) + self.mass_min
        self.massval = massval
        return massval
    
    def sample_sfr_prior(self, zval = 1.0, size=1):
        if self.sfr_prior_type == 'SFRflat':
            return np.random.uniform(size=size)*(self.sfr_max-self.sfr_min) + self.sfr_min
        elif self.sfr_prior_type == 'sSFRflat':
            return np.random.uniform(size=size)*(self.ssfr_max-self.ssfr_min) + self.ssfr_min + self.massval
        elif self.sfr_prior_type == 'sSFRlognormal':
            temp = np.random.lognormal(mean = self.ssfr_mean, sigma = self.ssfr_sigma, size=size) # about a ~2.5 dex width.
            temp = temp - np.exp(self.ssfr_mean) # centered at 0
            temp = np.log10(10.0/(cosmo.age(zval).value*1e9))-temp + self.ssfr_shift # and then at the sSFR at that redshift
            sfrval = temp + self.massval
            return sfrval
        else:
            print('unknown SFR prior type. choose from SFRflat, sSFRflat, or sSFRlognormal.')
            return np.nan

    def sample_tx_prior(self, size=1):
        if self.sfh_treatment == 'TNGlike':
            #tng_zvals = np.load('train_data/alpha_lookup_tables/tng_alpha_zvals.npy')
            #tng_alphas = np.load('train_data/alpha_lookup_tables/tng_alpha_Nparam_%.0f.npy' %self.Nparam)
            tng_zvals = np.load(get_file('train_data/alpha_lookup_tables','tng_alpha_zvals.npy'))
            tng_alphas = np.load(get_file('train_data/alpha_lookup_tables','tng_alpha_Nparam_%.0f.npy' %self.Nparam))
            tng_best_z_index = np.argmin(np.abs(tng_zvals - self.zval))
            self.tx_alpha = tng_alphas[0:,tng_best_z_index]

        if size == 1:
            temp_tx = np.cumsum(np.random.dirichlet(np.ones((self.Nparam+1,))*self.tx_alpha, size=size))[0:-1]
            return temp_tx
        else:
            temp_tx = np.cumsum(np.random.dirichlet(np.ones((self.Nparam+1,))*self.tx_alpha, size=size),axis=1)[0:,0:-1]
            return temp_tx

    def sample_z_prior(self, size=1):
        zval = np.random.uniform(size=size)*(self.z_max-self.z_min) + self.z_min
        self.zval = zval
        return zval

    def sample_Z_prior(self, size=1):
        if self.met_treatment == 'flat':
            return np.random.uniform(size=size)*(self.Z_max-self.Z_min) + self.Z_min
        elif self.met_treatment == 'massmet':
            met = np.random.normal(scale = self.massmet_width, size=size) + (self.massval - 7)/(10.8-7) - 1.0 # roughly solar met at MW-mass
            return met

    def sample_Av_prior(self, size=1):
        if self.dust_model == 'Calzetti':
            if self.dust_prior == 'flat':
                return np.random.uniform(size=size)*(self.Av_max-self.Av_min) + self.Av_min
            elif self.dust_prior == 'exp':
                return np.random.exponential(size=size)*(self.Av_exp_scale)
            else:
                print('unknown dust_prior. options are flat and exp')
        if self.dust_model == 'CF00':
            # using parameters from Pacifici+16 for now
            slope_ism = np.random.uniform(size=size)*(0.6) - 1.0 # ism slope in range [-1.0,-0.4]
            slope_bc = -1.3*np.ones((size,))
            mu_ISM_v = np.random.uniform(size=size)*0.6 + 0.1 #\mu in range [0.1,0.7]
            if self.dust_prior == 'flat':
                tau_v = (np.random.uniform(size=size)*(self.Av_max-self.Av_min) + self.Av_min)
            elif self.dust_prior == 'exp':
                tau_v = np.random.exponential(size=size)*(self.Av_exp_scale)
            else:
                print('unknown dust_prior. options are flat and exp')
            return slope_ism, slope_bc, mu_ISM_v, tau_v
        else:
            print('not currently coded up, please email me regarding this functionality.')
            return np.zeros(size=size)*np.nan

    def print_priors(self):
        print('--------------Priors:--------------')
        print('The prior on log mass is uniform from %.1f to %.1f.' %(self.mass_min, self.mass_max))
        if self.use_ssfr_prior == True:
            print('The prior on log sSFR_inst is -ve lognormal with mean %.1f and sigma %.1f.' %(self.ssfr_shift, self.ssfr_sigma))
        else:
            print('The prior on log SFR_inst is uniform from %.1f to %.1f.' %(self.sfr_min, self.sfr_max))
        print('The prior on tx is dirichlet with alpha = ',(self.tx_alpha))
        print('The prior on redshift is uniform from %.1f to %.1f.' %(self.z_min, self.z_max))
        print('The prior on log metallicity/Zsolar is uniform from %.1f to %.1f.' %(self.Z_min, self.Z_max))
        print('The prior on dust (model: '+self.dust_model+') is uniform with Av: %.1f to %.1f.' %(self.Av_min, self.Av_max))
        print('-----------------------------------')

    def sample_sfh_tuple(self):
        sfh_tuple = np.zeros((self.Nparam + 3,))
        sfh_tuple[0] = self.sample_mass_prior()
        sfh_tuple[1] = self.sample_sfr_prior()
        sfh_tuple[2] = self.Nparam
        sfh_tuple[3:] = self.sample_tx_prior()
        return sfh_tuple

    def sample_all_params(self, random_seed = np.random.randint(1)):
        np.random.seed(random_seed)
        temp_z = self.sample_z_prior()
        sfh_tuple = self.sample_sfh_tuple()
        temp_Z = self.sample_Z_prior()
        temp_Av = self.sample_Av_prior()
        return sfh_tuple, temp_Z, temp_Av, temp_z

    def make_N_prior_draws(self, size=10, random_seed = np.random.randint(1)):
        sfh_tuples = np.zeros((self.Nparam+3, size))
        Avs = np.zeros((size,))
        Zs = np.zeros((size,))
        zs = np.zeros((size,))
        for i in range(size):
            sfh_tuples[0:,i], Zs[i], Avs[i], zs[i] = self.sample_all_params(random_seed = random_seed+i*7)
        return sfh_tuples, Zs, Avs, zs

    def plot_prior_distributions(self, num_draws = 100000):

        a,b,c,d = self.make_N_prior_draws(size = num_draws, random_seed=10)
        theta_arr = np.vstack((a[0,0:], a[1,0:], a[3:,0:], b,c,d))

        txs = ['t'+'%.0f' %i for i in quantile_names(self.Nparam)]
        if self.dust_model == 'Calzetti':
            prior_labels = ['log M*', 'log SFR', 'Z', 'Av', 'z']
        elif self.dust_model == 'CF00':
            print('currently not implemented.')
            return
        prior_labels[2:2] = txs

        figure = corner.corner(theta_arr.T,labels=prior_labels,
                plot_datapoints=False, fill_contours=True,
                bins=50, smooth=1.0, levels=[1 - np.exp(-(1/1)**2/2),1 - np.exp(-(2/1)**2/2)],
                label_kwargs={"fontsize": 30})
        figure.subplots_adjust(right=1.5,top=1.5)
        plt.show()

    def plot_sfh_prior(self, numdraws = 100, ref_mstar = 10.0, zval = 5.606):

        sfhs = np.zeros((1000, numdraws))
        sfh_tuples = np.zeros((self.Nparam+3, numdraws))
        sfr_error = np.zeros((numdraws,))
        mass_error = np.zeros((numdraws,))

        for i in (range(numdraws)):
            sfh_tuples[0:,i], _,_,_ = self.sample_all_params(random_seed = i*7)
            ssfr = sfh_tuples[1,i] - sfh_tuples[0,i]
            sfh_tuples[0,i] = ref_mstar
            sfh_tuples[1,i] = ssfr + ref_mstar
            sfhs[0:,i], time = tuple_to_sfh(sfh_tuples[0:,i], zval = zval)
            sfr_error[i] = np.log10(sfhs[-1,i]) - sfh_tuples[1,i]
            mass_error[i] = np.log10(np.trapz(x=time*1e9, y=sfhs[0:,i])) - sfh_tuples[0,i]

        # sfr_error[sfr_error<-2] = -2
        # sfr_error[sfr_error>2] = 2
        # plt.plot(sfh_tuples[1,0:], sfr_error,'.')
        # plt.show()
        # plt.hist(sfr_error,30)
        # plt.show()
        #
        # plt.plot(sfh_tuples[0,0:], mass_error,'.')
        # plt.show()
        # plt.hist(mass_error,30)
        # plt.show()

        plt.figure(figsize=(12,6))
        plt.plot((np.amax(time)-time), np.nanmedian(sfhs,1),'k',lw=3)
        plt.fill_between((np.amax(time)-time), np.nanpercentile(sfhs,40,axis=1),
                         np.nanpercentile(sfhs,60,axis=1),color='k',alpha=0.1)
        plt.fill_between((np.amax(time)-time), np.nanpercentile(sfhs,25,axis=1),
                         np.nanpercentile(sfhs,75,axis=1),color='k',alpha=0.1)
        plt.fill_between((np.amax(time)-time), np.nanpercentile(sfhs,16,axis=1),
                         np.nanpercentile(sfhs,84,axis=1),color='k',alpha=0.1)
        plt.plot((np.amax(time)-time), sfhs[0:,0:6],'k',lw=1, alpha=0.3)
        plt.ylabel('normalized SFR')
        plt.xlabel('lookback time [Gyr]')
        plt.show()

        txs = ['t'+'%.0f' %i for i in quantile_names(self.Nparam)]
        pg_labels = ['log sSFR']
        pg_labels[1:1] = txs

        arr = np.vstack((sfh_tuples[1,0:]-10, sfh_tuples[3:,0:]))
        corner.corner(arr.T, labels=pg_labels, plot_datapoints=False, fill_contours=True,
                bins=50, smooth=1.0, levels=[1 - np.exp(-(1/1)**2/2),1 - np.exp(-(2/1)**2/2)], label_kwargs={"fontsize": 30})
        plt.subplots_adjust(right=1.5,top=1.5)
        plt.show()



import numpy as np
from tqdm import tqdm
import scipy.io as sio

try:
    import emcee
    from scipy.interpolate import NearestNDInterpolator
except:
    print('running without emcee')

from .priors import *
from .gp_sfh import *
from .plotter import *
from .sed_fitter import *

def get_mcmc_variables(atlas):

    model_params = np.vstack((atlas['mstar'],atlas['sfr'],atlas['sfh_tuple'][0:,-3:].T,atlas['met'].ravel(),atlas['dust'].ravel(),atlas['zval'].ravel())).T
    model_seds = atlas['sed']
    interp_engine = NearestNDInterpolator(model_params, model_seds)
    param_limits = [np.amin(model_params,0), np.amax(model_params,0)]
    param_limits[0][1] = -3.0
    param_limits[0][2] = 0.0

    return model_params, model_seds, interp_engine, param_limits

# def log_prior(theta):
#
#     # change these to match the priors i'm using, esp. for SFHs
#
#     mass, sfr, t25, t50, t75, met, dust, zval = theta
#     if mass_min < mass < mass_max and sfr_min < sfr < sfr_max \
#     and 0.0 < t25 < t50 and t25 < t50 < t75 and t50 < t75 < t75_max \
#     and dust_min < dust < dust_max and met_min < met < met_max \
#     and zval_min < zval < zval_max:
#         return 0.0
#     return -np.inf

def log_prior(theta, param_limits):

    # change these to match the priors i'm using, esp. for SFHs
    mass, sfr, t25, t50, t75, met, dust, zval = theta

    if (param_limits[0] < theta).all() and (param_limits[1] > theta).all() and (0<t25<t50) and (t25<t50<t75):
        return 0.0
    return -np.inf

def log_likelihood(theta, sed, sed_err, interp_engine):

    pred_sed = interp_engine(theta)[0]
    fit_mask = (sed > 0)

    chi2 = (pred_sed[fit_mask] - sed[fit_mask])**2 / (sed_err[fit_mask])**2

    return -0.5 * np.sum(chi2)

def log_probability(theta, sed, sed_err, interp_engine, param_limits):
    lp = log_prior(theta, param_limits)
    if not np.isfinite(lp):
        return -np.inf
    return lp + log_likelihood(theta, sed, sed_err, interp_engine)

def run_emceesampler(gal_sed, gal_err, atlas, fit_mask=[], zbest=None, deltaz=None, nwalkers = 100, epochs = 1000, plot_posteriors = False):

    model_params, model_seds, interp_engine, param_limits = get_mcmc_variables(atlas)

    pos = model_params[0:nwalkers,0:].copy()
    pos[pos[0:,1]<-2,1] = -2
    _, ndim = pos.shape

    _, norm_fac = evaluate_sed_likelihood(gal_sed,gal_err,atlas,fit_mask=[],
                                            zbest=None,deltaz=None)

    sampler = emcee.EnsembleSampler(nwalkers, ndim, log_probability,
                                    args=(gal_sed/norm_fac, gal_err/norm_fac, interp_engine, param_limits))
    sampler.run_mcmc(pos, epochs, progress=True)

    if plot_posteriors == True:
        plot_emcee_posterior(sampler, norm_fac)

    return sampler, norm_fac

def plot_emcee_posterior(sampler, norm_fac, discard = 100, thin = 1):

    flat_samples = sampler.get_chain(discard=discard, thin=thin, flat=True)
    flat_samples[0:,0] = flat_samples[0:,0] + np.log10(norm_fac)
    flat_samples[0:,1] = flat_samples[0:,1] + np.log10(norm_fac)

    atlas_labels = ['log M*', 'log SFR', 't25','t50','t75', 'Z', 'Av', 'z']

    fig = corner.corner(
        flat_samples, labels = atlas_labels,
        plot_datapoints=False, fill_contours=True,
        bins=20, smooth=1.0,
        quantiles=(0.16, 0.84), levels=[1 - np.exp(-(1/1)**2/2),1 - np.exp(-(2/1)**2/2)],
        label_kwargs={"fontsize": 30}, show_titles=True
    );
    fig.subplots_adjust(right=1.5,top=1.5)
    plt.show()
